! ******************************************************* !
!
subroutine G_ab_RPT(k, eta1, eta2, G11, G12, G21, G22)
!                                                         !
!                                                         !
!             Time-stamp: <2011-03-01 10:37:35 t-okamura> ! 
! ******************************************************* !
!
!     Renormalized propagator by Crocce & Scoccimarro (2006b)
!
!     computing correction factor "alpha(z)" has been included
!
  use data,only : ikmax, ik_max, ak, ialpha, ff, gg, hh, ii, iloop
  implicit none

  double precision, intent(in) :: k, eta1, eta2
  double precision, intent(out) :: G11, G12, G21, G22

  integer j, jmin, jmax, m
  double precision eta, XX, YY, dd2, A, B
  double precision alpha, beta, gamma, delta
  double precision beta_d, beta_g, gamma_d, gamma_g
  double precision fk, gk, hk, ik, s, ds
  double precision afk, dfk, agk, dgk, ghk, gik
  double precision alpha_z
! -------------------------------------------------

! //////// interpolating one-loop propagator ////////

  call hunt(ak, ik_max, k, j)

  jmin = max(1, j - 2)
  jmax = min(ik_max, j + 2)
  m = jmax-jmin+1

  call polint(ak(jmin:jmax), ff(jmin:jmax), m, k, s, ds)
  fk = s
  call polint(ak(jmin:jmax), gg(jmin:jmax), m, k, s, ds)
  gk = s
  call polint(ak(jmin:jmax), hh(jmin:jmax), m, k, s, ds)
  hk = s
  call polint(ak(jmin:jmax), ii(jmin:jmax), m, k, s, ds)
  ik = s

! //////// interpolating alpha ////////

  if(ialpha == 0) then 
     call find_alpha(eta1, alpha_z)
     alpha_z = alpha_z * alpha_z
  else
     alpha_z = 1.d0
  endif

! ///////// computing growth functions ////////

  eta = eta1 - eta2

  call growth_functions(eta, alpha, beta, gamma, delta)

  A = exp(eta)
  B = exp(1.5d0*eta)

  beta_d = beta * B
  beta_g = beta / A
  gamma_d = gamma * B
  gamma_g = gamma / A

! ///////// renormalized propagators ////////
  afk = alpha   * fk
  dfk = delta   * fk
  agk = alpha   * gk
  dgk = delta   * gk
  ghk = gamma_d * hk
  gik = gamma_g * ik
  
  dd2 = exp(2.d0*eta2)

  XX = dd2 * (afk - beta_g * ik)
  YY = dd2 * (dgk - ghk)
  G11 = 0.6d0 * A * exp(alpha_z*XX) + 0.4d0 / B * exp(alpha_z*YY)

  XX = dd2 * (afk - beta_g * hk)
  YY = dd2 * (dfk - ghk)
  G12 = 0.4d0 * A * exp(alpha_z*XX) - 0.4d0 / B * exp(alpha_z*YY)

  if(iloop == 1)then
     XX = dd2 * (agk + gik)
     YY = dd2 * (dgk + beta_d * ik)
  elseif(iloop == 2)then
     XX = dd2 * (agk + gamma_g * hk)
     YY = dd2 * (dgk + gamma_d * ik)
  endif
  G21 = 0.6d0 * A * exp(alpha_z*XX) - 0.6d0 / B * exp(alpha_z*YY)

  XX = dd2 * (agk - 1.5d0 * gik)
  YY = dd2 * (dfk - beta_d * hk / 1.5d0)
  G22 = 0.4d0 * A * exp(alpha_z*XX) + 0.6d0 / B * exp(alpha_z*YY)
  return
end subroutine G_ab_RPT
! ******************************************************* !
!
subroutine G_ab_Closure(k, eta1, eta2, G11, G12, G21, G22)
!
! ******************************************************* !
!
!     Renormalized propagator by closure theory
!
  use data,only : ikmax, ik_max, ak, ff, gg, hh, ii, ialpha, iloop
  implicit none

  double precision, intent(in) :: k, eta1, eta2
  double precision, intent(out) :: G11, G12, G21, G22
  integer j, jmin, jmax, m
  double precision eta, XX, YY, dd, bessj1, A, B
  double precision alpha, beta, gamma, delta
  double precision beta_d, beta_g, gamma_d, gamma_g
  double precision fk, gk, hk, ik, s, ds
  double precision afk, dfk, agk, dgk, ghk, gik
  double precision alpha_z
! -------------------------------------------------

! //////// interpolating one-loop propagator ////////

  call hunt(ak, ik_max, k, j)
  
  jmin = max(1, j - 2)
  jmax = min(ik_max, j + 2)
  m = jmax-jmin+1

  call polint(ak(jmin:jmax), ff(jmin:jmax), m, k, s, ds)
  fk = s
  call polint(ak(jmin:jmax), gg(jmin:jmax), m, k, s, ds)
  gk = s
  call polint(ak(jmin:jmax), hh(jmin:jmax), m, k, s, ds)
  hk = s
  call polint(ak(jmin:jmax), ii(jmin:jmax), m, k, s, ds)
  ik = s

! //////// interpolating alpha ////////

  if(ialpha == 0) then 
     call find_alpha(eta1, alpha_z)
  else
     alpha_z = 1.d0
  endif

!     ///////// computing growth functions ////////

  eta = eta1 - eta2

  call growth_functions(eta, alpha, beta, gamma, delta)

  A = exp(eta)
  B = exp(1.5d0*eta)

  beta_d = beta * B
  beta_g = beta / A
  gamma_d = gamma * B
  gamma_g = gamma / A

! ///////// renormalized propagators ////////
  afk = alpha   * fk
  dfk = delta   * fk
  agk = alpha   * gk
  dgk = delta   * gk
  ghk = gamma_d * hk
  gik = gamma_g * ik

  dd = exp(eta2)
  dd = dd * alpha_z
  
  XX = dd * sqrt(2.d0 * abs(afk - beta_g * ik))
  YY = dd * sqrt(2.d0 * abs(dgk - ghk))     
  G11 = 0.6d0 * A * bessj1(2.d0*XX)/XX + 0.4d0 / B * bessj1(2.d0*YY)/YY 

  XX = dd * sqrt(2.d0 * abs(afk - beta_g * hk))
  YY = dd * sqrt(2.d0 * abs(dfk - ghk))
  G12 = 0.4d0 * A * bessj1(2.d0*XX)/XX - 0.4d0 / B * bessj1(2.d0*YY)/YY

  if(iloop == 1)then
     XX = dd * sqrt(2.d0 * abs(agk + gik))
     YY = dd * sqrt(2.d0 * abs(dgk + beta_d * ik))
  elseif(iloop == 2)then
     XX = dd * sqrt(2.d0 * abs(alpha * gk + gamma_g * hk))
     YY = dd * sqrt(2.d0 * abs(delta * gk + gamma_d * ik))
  endif
  G21 = 0.6d0 * A * bessj1(2.d0*XX)/XX - 0.6d0 / B * bessj1(2.d0*YY)/YY

  XX = dd * sqrt(2.d0 * abs(agk - 1.5d0 * gik))
  YY = dd * sqrt(2.d0 * abs(dfk - beta_d * hk / 1.5d0 ))
  G22 = 0.4d0 * A* bessj1(2.d0*XX)/XX + 0.6d0 / B * bessj1(2.d0*YY)/YY
  return
end subroutine G_ab_Closure
! ******************************************************* !
!
subroutine G_ab_linear(eta1, eta2, G11, G12, G21, G22)
!
! ******************************************************* !
!
!     Linear propagator 
!
  implicit none

  double precision, intent(in) :: eta1, eta2
  double precision, intent(out) :: G11, G12, G21, G22
  double precision eta, A, B
! -------------------------------------------------
   
  eta = eta1 - eta2
  A = exp(eta)
  B = exp(-1.5d0*eta)

  G11 = 0.6d0 * A + 0.4d0 * B 
  G12 = 0.4d0 * A - 0.4d0 * B
  G21 = 0.6d0 * A - 0.6d0 * B 
  G22 = 0.4d0 * A + 0.6d0 * B
  
  return
end subroutine G_ab_linear
! ******************************************************* !
!
subroutine growth_functions(eta, alpha, beta, gamma, delta)
!
! ******************************************************* !
!
  implicit none

  double precision, intent(in) :: eta
  double precision, intent(out) :: alpha, beta, gamma, delta
  double precision A, B, C, D, E
! ------------------------------------------

  A = exp(eta)
  B = sqrt(A)        ! exp(0.5eta)
  C = 1.d0/A/B       ! exp(-1.5eta)
  D = A * A          ! exp(2eta)
  E = D * A * B      ! exp(3.5eta)

  alpha = D - 1.4d0 * A + 0.4d0 * C
  beta  = 0.6d0 * D - A + 0.4d0 / B
  gamma = 0.4d0 * D - B + 0.6d0 / B
  delta = 0.4d0 * E - 1.4d0 * A + 1.d0

  return
end subroutine growth_functions
! ******************************************************* !
!
subroutine one_loop_propagator
!
! ******************************************************* !
!
!     computing one-loop propagator 
! 
  use data, only : pi, ikmax, ik_max, ak, pk, ff, gg, hh, ii
  implicit none
  
  integer, parameter :: ix_max = 6000
  integer  ik, ix
  double precision kernel_ff, kernel_gg, kernel_hh, kernel_ii
  double precision kmin, kmax, xmin, xmax
  double precision pklin1, pklin2
  double precision x1, x2, k, fact, dx
  external kernel_ff, kernel_gg, kernel_hh, kernel_ii
! -------------------------------------------------

  kmin = ak(1)
  kmax = ak(ik_max)

  ff(:) = 0.d0
  gg(:) = 0.d0
  hh(:) = 0.d0
  ii(:) = 0.d0

  do ik = 1, ik_max    
     k = ak(ik)
     
     xmin = kmin / k
     xmax = kmax / k

     x1 = xmin
     call find_pk(k*x1, pklin1)

     do ix = 1, ix_max-1
        
        x2 = xmin * (xmax/xmin) ** (dble(ix)/dble(ix_max-1))
        call find_pk(k*x2, pklin2)
        dx = (x2-x1)*0.5d0

        ff(ik) = ff(ik)                                             &
             &  + (kernel_ff(x1) * pklin1 + kernel_ff(x2) * pklin2) &
             &  * dx
        gg(ik) = gg(ik)                                             &
             &  + (kernel_gg(x1) * pklin1 + kernel_gg(x2) * pklin2) &
             &  * dx
        hh(ik) = hh(ik)                                             &
             &  + (kernel_hh(x1) * pklin1 + kernel_hh(x2) * pklin2) &
             &  * dx
        ii(ik) = ii(ik)                                             &
             &  + (kernel_ii(x1) * pklin1 + kernel_ii(x2) * pklin2) &
             &  * dx
    
        x1 = x2
        pklin1 = pklin2
     enddo
     
     fact = k ** 3/ (2.d0*pi*pi) 
     ff(ik) = fact * ff(ik) 
     gg(ik) = fact * gg(ik) 
     hh(ik) = fact * hh(ik) 
     ii(ik) = fact * ii(ik) 
  enddo
  return
end subroutine one_loop_propagator
! ******************************************************* !
!
function kernel_ff(x)
!
! ******************************************************* !
  implicit none
  double precision, intent(in) :: x
  double precision kernel_ff, x2

  if( (x <= 10.d0) .and. (abs(x-1.d0) >= 1d-2) ) then
     x2 = x * x
     kernel_ff = 6.d0/x2 - 79.d0 + 50.d0*x2 - 21.d0*x2*x2  & 
          &  + 0.75d0*(1.d0/x-x)**3 * (2.d0 + 7.d0*x2)        &
          &  * log(abs((1.d0-x)/(1.d0+x))**2)
     kernel_ff = kernel_ff / 504.d0
  elseif(abs(x-1.d0) < 1d-2) then 
     kernel_ff = - 11.d0/126.d0 + (x-1.d0)/126.d0  & 
          &      - 29.d0/252.d0 * (x-1.d0)**2
  elseif(x > 10.d0) then
     x2 = x * x
     kernel_ff = - 61.d0/630.d0 + 2.d0/105.d0/x2  &
          &     - 10.d0/1323.d0/x2/x2
  endif
  return
end function kernel_ff
! ******************************************************* !
!
function kernel_gg(x)
!
! ******************************************************* !

  implicit none
  double precision, intent(in) :: x
  double precision kernel_gg, x2

  if( (x <= 10.d0) .and. (abs(x-1.d0) >= 1d-2) ) then
     x2 = x * x
     kernel_gg = 6.d0/x2 - 41.d0 + 2.d0*x2 - 3.d0*x2*x2  &
          &   + 0.75d0*(1.d0/x-x)**3 * (2.d0 + x2)          &
          &   * log(abs((1.d0-x)/(1.d0+x))**2)
     kernel_gg = kernel_gg / 168.d0
  elseif(abs(x-1.d0) < 1d-2) then 
     kernel_gg = - 3.d0/14.d0 - 5.d0/42.d0*(x-1.d0) &
          &   +  (x-1.d0)**2/84.d0
  elseif(x > 10.d0) then
     x2 = x * x
     kernel_gg = - 0.3d0 + 26.d0/245.d0/x2   &
          &      - 38.d0/2205.d0/x2/x2
  endif
  return
end function kernel_gg
! ******************************************************* !
!
function kernel_hh(x)
!
! ******************************************************* !

  implicit none
  double precision, intent(in) :: x
  double precision kernel_hh, x2

  if( (x <= 10.d0) .and. (abs(x-1.d0) >= 1d-3) ) then
     x2 = x * x
     kernel_hh = 6.d0/x2 + 1.d0 + 9.d0*x2*x2                      &  
          &  + 0.75d0*(1.d0/x-x)**2 * (2.d0/x + 5.d0*x + 3.d0*x2*x) &
          &  * log(abs((1.d0-x)/(1.d0+x))**2)
     kernel_hh = kernel_hh / 24.d0
  elseif(abs(x-1.d0) < 1d-3) then 
     kernel_hh = 2.d0/3.d0 + (x-1.d0)
  elseif(x > 10.d0) then
     x2 = x * x
     kernel_hh = 19.d0/30.d0 + 32.d0/105.d0/x2  &
          &    - 4.d0/21.d0 /x2/x2
  endif
  return
end function kernel_hh
! ******************************************************* !
!
function kernel_ii(x)
!
! ******************************************************* !

  implicit none
  double precision, intent(in) :: x
  double precision kernel_ii, x2

  if( (x <= 10.d0) .and. (abs(x-1.d0) >= 1d-3) ) then
     x2 = x * x
     kernel_ii = 6.d0/x2 + 29.d0 - 18.d0*x2 + 27.d0*x2*x2          &
          &    + 0.75d0*(1.d0/x-x)**2 * (2.d0/x + 9.d0*x + 9.d0*x2*x) &
          &    * log(abs((1.d0-x)/(1.d0+x))**2)
     kernel_ii = - kernel_ii / 72.d0
  elseif(abs(x-1.d0) < 1d-3) then 
     kernel_ii = -11.d0/18.d0 - 5.d0/6.d0*(x-1.d0)
  elseif(x > 10.d0) then
     x2 = x * x
     kernel_ii = -67.d0/90.d0 + 2.d0/315.d0/x2  &
          &     + 26.d0/315.d0 /x2/x2
  endif
  return
end function kernel_ii
! ******************************************************* !
!
FUNCTION bessj1(x)
!
! ******************************************************* !

  implicit none
  double precision, intent(in) :: x
  double precision bessj1
  double precision ax, xx, z, y
  double precision, parameter :: p(5) &
       & = (/ 1.d0, 0.183105d-2, -0.3516396496d-4, &
       &      0.2457520174d-5, -0.240337019d-6 /)
  double precision, parameter :: q(5) &
       & = (/ 0.04687499995d0, -0.2002690873d-3, &
       &      0.8449199096d-5, -0.88228987d-6, 0.105787412d-6 /)
  double precision, parameter :: r(6) &
       & = (/ 72362614232.d0, -7895059235.d0,  242396853.1d0, &
       &      -2972611.439d0,  15704.48260d0, -30.16036606d0 /)
  double precision, parameter :: s(6) &
       & = (/ 144725228442.d0, 2300535178.d0,  18583304.74d0, &
       &       99447.43394d0,  376.9991397d0,  1.d0 /)
  
  if(abs(x) < 8.d0)then
     y = x * x
     bessj1 = x*(r(1)+y*(r(2)+y*(r(3)+y*(r(4)+y*(r(5)+y*r(6)))))) &
          & /(s(1)+y*(s(2)+y*(s(3)+y*(s(4)+y*(s(5)+y*s(6))))))
  else
     ax = abs(x)
     z = 8.d0/ax
     y = z * z
     xx = ax-2.356194491d0
     bessj1 = sqrt(0.636619772d0/ax) &
          & *(cos(xx)*(p(1)+y*(p(2)+y*(p(3)+y*(p(4)+y*p(5))))) &
          & -z*sin(xx)*(q(1)+y*(q(2)+y*(q(3)+y*(q(4)+y*q(5))))))*sign(1.d0,x)
  endif
  return
END FUNCTION bessj1
!  (C) Copr. 1986-92 Numerical Recipes Software 
